from os.path import join

from cobra.io import read_sbml_model, write_sbml_model


def _check_sbml_annotations(model):
    """Checks the annotations from the annotation.xml. """
    assert model is not None

    # model annotation
    # {'bigg.model': 'e_coli_core', 'doi': '10.1128/ecosalplus.10.2.1',
    #  'taxonomy': '511145'}
    annotation = model.annotation
    assert annotation is not None
    assert len(annotation) == 3
    for key in ["bigg.model", "doi", "taxonomy"]:
        assert key in annotation
    assert annotation["bigg.model"] == "e_coli_core"
    assert annotation["doi"] == "10.1128/ecosalplus.10.2.1"
    assert annotation["taxonomy"] == "511145"

    # gene annotation
    # {'asap': 'ABE-0006162', 'ncbigene': '946368', 'uniprot': 'P33221',
    #  'ncbigi': 'gi:16129802', 'ecogene': 'EG11809'}
    annotation = model.genes.G1.annotation
    assert len(annotation) == 5
    for key in ["asap", "ncbigene", "uniprot", "ncbigi", "ecogene"]:
        assert key in annotation
    assert annotation["asap"] == "ABE-0006162"
    assert annotation["ncbigene"] == "946368"
    assert annotation["uniprot"] == "P33221"
    assert annotation["ncbigi"] == "gi:16129802"
    assert annotation["ecogene"] == "EG11809"

    # compartment annotation
    # FIXME: add tests with first class compartment model
    # annotation = model.compartments.c.annotation
    # assert len(annotation) == 1
    # for key in ["bigg.compartment"]:
    #    assert key in annotation
    # assert annotation["bigg.compartment"] == "c"

    # metabolite/species annotation
    # {'inchi': 'InChI=1S/C3H8O2/c1-3(5)2-4/h3-5H,2H2,1H3/t3-/m0/s1',
    # 'bigg.metabolite': '13dpg', 'chebi': ['CHEBI:11881', 'CHEBI:16001',
    # 'CHEBI:1658', 'CHEBI:20189', 'CHEBI:57604'],
    # 'metanetx.chemical': 'MNXM261',
    # 'kegg.compound': ['C00236', 'C02917'],
    # 'seed.compound': 'cpd00203', 'hmdb': ['HMDB62758', 'HMDB06213'],
    # 'biocyc': 'META:DPG'}
    annotation = model.metabolites.A.annotation
    for key in ["inchi", "bigg.metabolite", "chebi", "metanetx.chemical",
                "kegg.compound", "seed.compound", "hmdb", "biocyc"]:
        assert key in annotation
    assert annotation[
               "inchi"] == "InChI=1S/C3H8O2/c1-3(5)2-4/h3-5H,2H2,1H3/t3-/m0/s1"  # noqa: E501

    # reaction annotation
    # {'kegg.reaction': 'R00228', 'sbo': 'SBO:0000375',
    # 'ec-code': '1.2.1.10', 'rhea': ['23288', '23289', '23290', '23291'],
    # 'metanetx.reaction': 'MNXR95210', 'bigg.reaction': 'ACALD',
    # 'biocyc': 'META:ACETALD-DEHYDROG-RXN'}
    annotation = model.reactions.R1.annotation
    for key in ["kegg.reaction", "sbo", "ec-code", "rhea",
                "metanetx.reaction", "bigg.reaction", "biocyc"]:
        assert key in annotation
    assert annotation["biocyc"] == 'META:ACETALD-DEHYDROG-RXN'


def test_read_sbml_annotations(data_directory):
    """Test reading and writing annotations."""
    with open(join(data_directory, "annotation.xml"), "r") as f_in:
        model1 = read_sbml_model(f_in)
        _check_sbml_annotations(model1)


def test_read_write_sbml_annotations(data_directory, tmp_path):
    """Test reading and writing annotations."""
    with open(join(data_directory, "annotation.xml"), "r") as f_in:
        model1 = read_sbml_model(f_in)

    sbml_path = join(str(tmp_path), "test.xml")
    with open(sbml_path, "w") as f_out:
        write_sbml_model(model1, f_out)

    with open(sbml_path, "r") as f_in:
        model2 = read_sbml_model(f_in)
        _check_sbml_annotations(model2)



# -*- coding: utf-8 -*-

"""Test functionalities of MetaboliteSummary."""

from __future__ import absolute_import

import numpy as np
import pytest

from cobra.flux_analysis.parsimonious import pfba
from cobra.test.test_core.test_summary import (
    captured_output, check_in_line, check_line)


@pytest.mark.parametrize("met", ["q8_c"])
def test_metabolite_summary_to_table_previous_solution(model, opt_solver, met):
    """Test metabolite summary._to_table() of previous solution."""
    model.solver = opt_solver
    solution = pfba(model)

    expected_entry = ['PRODUCING CYTBD     100    43.6  2.0 h_c + 0.5 o2_c + '
                      'q8h2_c --> h2o_c + 2.0 h_']

    with captured_output() as (out, _):
        print(model.metabolites.get_by_id(met).summary(solution))
        check_in_line(out.getvalue(), expected_entry)


@pytest.mark.parametrize("met", ["q8_c"])
def test_metabolite_summary_to_frame_previous_solution(model, opt_solver, met):
    """Test metabolite summary.to_frame() of previous solution."""
    model.solver = opt_solver
    solution = pfba(model)

    expected_percent = [100.0, 88.4, 11.6]

    out_df = model.metabolites.get_by_id(met).summary(solution).to_frame()

    assert out_df['PERCENT'].round(1).tolist() == expected_percent


@pytest.mark.parametrize("met, names", [
    ("q8_c", False),
    ("q8_c", True)
])
def test_metabolite_summary_to_table(model, opt_solver, met, names):
    """Test metabolite summary._to_table()."""
    model.solver = opt_solver
    model.optimize()

    if names:
        expected_entry = ['PRODUCING cytochrome oxidase bd '
                          '(ubiquinol-8: 2 protons)    100    43.6  '
                          '2.0 H+ + 0.5 O2 + Ubiquinol-8 --> H2O + 2.0 H+']
    else:
        expected_entry = ['PRODUCING CYTBD     100    43.6  2.0 h_c + '
                          '0.5 o2_c + q8h2_c --> h2o_c + 2.0 h_']

    with captured_output() as (out, _):
        print(model.metabolites.get_by_id(met).summary(names=names))
        check_in_line(out.getvalue(), expected_entry)


@pytest.mark.parametrize("met, names", [
    ("q8_c", False),
    ("q8_c", True)
])
def test_metabolite_summary_to_frame(model, opt_solver, met, names):
    """Test metabolite summary.to_frame()."""
    model.solver = opt_solver
    model.optimize()

    expected_percent = [100.0, 88.4, 11.6]

    out_df = model.metabolites.get_by_id(met).summary(names=names).to_frame()

    assert out_df['PERCENT'].round(1).tolist() == expected_percent


@pytest.mark.parametrize("fraction, met", [(0.99, "fdp_c")])
def test_metabolite_summary_to_table_with_fva(model, opt_solver, fraction,
                                              met):
    """Test metabolite summary._to_table() (using FVA)."""
    #     pytest.xfail("FVA currently buggy")

    model.solver = opt_solver
    model.optimize()

    expected_entry = ['PRODUCING PFK   100     7.48    6.17      9.26    '
                      'atp_c + f6p_c --> adp_c + fdp_c + h_c']

    with captured_output() as (out, _):
        print(model.metabolites.get_by_id(met).summary(fva=fraction))
        check_line(out.getvalue(), expected_entry)


@pytest.mark.parametrize("fraction, met", [(0.99, "fdp_c")])
def test_metabolite_summary_to_frame_with_fva(model, opt_solver, fraction,
                                              met):
    """Test metabolite summary.to_frame() (using FVA)."""

    model.solver = opt_solver
    model.optimize()

    expected_percent = [100.0, 100.0, 0.0]

    out_df = model.metabolites.get_by_id(met).summary(fva=fraction).to_frame()

    assert np.allclose(
        out_df['PERCENT'].tolist(),
        expected_percent,
        1e-6, 1e-6)



# -*- coding: utf-8 -*-

"""Test functions of configuration.py"""

from __future__ import absolute_import

from cobra.core import Configuration, Model
from cobra.util.solver import interface_to_str


def test_default_bounds():
    """Verify the default bounds."""
    config = Configuration()
    assert config.bounds == (-1000.0, 1000.0)


def test_bounds():
    """Test changing bounds."""
    config = Configuration()
    config.bounds = 100.0, 10000.0
    assert config.bounds == (100.0, 10000.0)
    # Restore default values.
    config.bounds = -1000.0, 1000.0


def test_solver():
    """Test assignment of different solvers."""
    config = Configuration()
    config.solver = "glpk"
    assert interface_to_str(config.solver) == "glpk"
    config.solver = "glpk_exact"
    assert interface_to_str(config.solver) == "glpk_exact"
    # Restore default solver.
    config.solver = "glpk"


def test_default_tolerance(model):
    """Verify the default solver tolerance."""
    config = Configuration()
    config.solver = "glpk"
    assert config.tolerance == 1e-07
    # Test the consistency between cobra.core.Configuration.tolerance and
    # cobra.core.Model.tolerance
    assert config.tolerance == model.tolerance


def test_toy_model_tolerance_with_different_default():
    """Verify that different default tolerance is respected by Model."""
    config = Configuration()
    config.tolerance = 1e-05

    toy_model = Model(name="toy model")
    assert toy_model.tolerance == 1e-05


def test_tolerance_assignment(model):
    """Test assignment of solver tolerance."""
    model.tolerance = 1e-06
    assert model.tolerance == 1e-06



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import logging
import multiprocessing
from builtins import map
from warnings import warn

from numpy import zeros
from optlang.symbolics import Zero
from pandas import DataFrame

from cobra.core import Configuration, get_solution
from cobra.flux_analysis.deletion import (
    single_gene_deletion, single_reaction_deletion)
from cobra.flux_analysis.helpers import normalize_cutoff
from cobra.flux_analysis.loopless import loopless_fva_iter
from cobra.flux_analysis.parsimonious import add_pfba
from cobra.util import solver as sutil


LOGGER = logging.getLogger(__name__)
CONFIGURATION = Configuration()


def _init_worker(model, loopless, sense):
    """Initialize a global model object for multiprocessing."""
    global _model
    global _loopless
    _model = model
    _model.solver.objective.direction = sense
    _loopless = loopless


def _fva_step(reaction_id):
    global _model
    global _loopless
    rxn = _model.reactions.get_by_id(reaction_id)
    # The previous objective assignment already triggers a reset
    # so directly update coefs here to not trigger redundant resets
    # in the history manager which can take longer than the actual
    # FVA for small models
    _model.solver.objective.set_linear_coefficients(
        {rxn.forward_variable: 1, rxn.reverse_variable: -1}
    )
    _model.slim_optimize()
    sutil.check_solver_status(_model.solver.status)
    if _loopless:
        value = loopless_fva_iter(_model, rxn)
    else:
        value = _model.solver.objective.value
    # handle infeasible case
    if value is None:
        value = float("nan")
        LOGGER.warning(
            "Could not get flux for reaction %s,  setting "
            "it to NaN. This is usually due to numerical instability.",
            rxn.id,
        )
    _model.solver.objective.set_linear_coefficients(
        {rxn.forward_variable: 0, rxn.reverse_variable: 0}
    )
    return reaction_id, value


def flux_variability_analysis(
    model,
    reaction_list=None,
    loopless=False,
    fraction_of_optimum=1.0,
    pfba_factor=None,
    processes=None,
):
    """
    Determine the minimum and maximum possible flux value for each reaction.

    Parameters
    ----------
    model : cobra.Model
        The model for which to run the analysis. It will *not* be modified.
    reaction_list : list of cobra.Reaction or str, optional
        The reactions for which to obtain min/max fluxes. If None will use
        all reactions in the model (default).
    loopless : boolean, optional
        Whether to return only loopless solutions. This is significantly
        slower. Please also refer to the notes.
    fraction_of_optimum : float, optional
        Must be <= 1.0. Requires that the objective value is at least the
        fraction times maximum objective value. A value of 0.85 for instance
        means that the objective has to be at least at 85% percent of its
        maximum.
    pfba_factor : float, optional
        Add an additional constraint to the model that requires the total sum
        of absolute fluxes must not be larger than this value times the
        smallest possible sum of absolute fluxes, i.e., by setting the value
        to 1.1 the total sum of absolute fluxes must not be more than
        10% larger than the pFBA solution. Since the pFBA solution is the
        one that optimally minimizes the total flux sum, the ``pfba_factor``
        should, if set, be larger than one. Setting this value may lead to
        more realistic predictions of the effective flux bounds.
    processes : int, optional
        The number of parallel processes to run. If not explicitly passed,
        will be set from the global configuration singleton.

    Returns
    -------
    pandas.DataFrame
        A data frame with reaction identifiers as the index and two columns:
        - maximum: indicating the highest possible flux
        - minimum: indicating the lowest possible flux

    Notes
    -----
    This implements the fast version as described in [1]_. Please note that
    the flux distribution containing all minimal/maximal fluxes does not have
    to be a feasible solution for the model. Fluxes are minimized/maximized
    individually and a single minimal flux might require all others to be
    suboptimal.

    Using the loopless option will lead to a significant increase in
    computation time (about a factor of 100 for large models). However, the
    algorithm used here (see [2]_) is still more than 1000x faster than the
    "naive" version using ``add_loopless(model)``. Also note that if you have
    included constraints that force a loop (for instance by setting all fluxes
    in a loop to be non-zero) this loop will be included in the solution.

    References
    ----------
    .. [1] Computationally efficient flux variability analysis.
       Gudmundsson S, Thiele I.
       BMC Bioinformatics. 2010 Sep 29;11:489.
       doi: 10.1186/1471-2105-11-489, PMID: 20920235

    .. [2] CycleFreeFlux: efficient removal of thermodynamically infeasible
       loops from flux distributions.
       Desouki AA, Jarre F, Gelius-Dietrich G, Lercher MJ.
       Bioinformatics. 2015 Jul 1;31(13):2159-65.
       doi: 10.1093/bioinformatics/btv096.
    """
    if reaction_list is None:
        reaction_ids = [r.id for r in model.reactions]
    else:
        reaction_ids = [
            r.id for r in model.reactions.get_by_any(reaction_list)
        ]

    if processes is None:
        processes = CONFIGURATION.processes
    num_reactions = len(reaction_ids)
    processes = min(processes, num_reactions)

    fva_result = DataFrame(
        {
            "minimum": zeros(num_reactions, dtype=float),
            "maximum": zeros(num_reactions, dtype=float),
        },
        index=reaction_ids,
    )
    prob = model.problem
    with model:
        # Safety check before setting up FVA.
        model.slim_optimize(
            error_value=None,
            message="There is no optimal solution for the "
            "chosen objective!",
        )
        # Add the previous objective as a variable to the model then set it to
        # zero. This also uses the fraction to create the lower/upper bound for
        # the old objective.
        # TODO: Use utility function here (fix_objective_as_constraint)?
        if model.solver.objective.direction == "max":
            fva_old_objective = prob.Variable(
                "fva_old_objective",
                lb=fraction_of_optimum * model.solver.objective.value,
            )
        else:
            fva_old_objective = prob.Variable(
                "fva_old_objective",
                ub=fraction_of_optimum * model.solver.objective.value,
            )
        fva_old_obj_constraint = prob.Constraint(
            model.solver.objective.expression - fva_old_objective,
            lb=0,
            ub=0,
            name="fva_old_objective_constraint",
        )
        model.add_cons_vars([fva_old_objective, fva_old_obj_constraint])

        if pfba_factor is not None:
            if pfba_factor < 1.0:
                warn(
                    "The 'pfba_factor' should be larger or equal to 1.",
                    UserWarning,
                )
            with model:
                add_pfba(model, fraction_of_optimum=0)
                ub = model.slim_optimize(error_value=None)
                flux_sum = prob.Variable("flux_sum", ub=pfba_factor * ub)
                flux_sum_constraint = prob.Constraint(
                    model.solver.objective.expression - flux_sum,
                    lb=0,
                    ub=0,
                    name="flux_sum_constraint",
                )
            model.add_cons_vars([flux_sum, flux_sum_constraint])

        model.objective = Zero  # This will trigger the reset as well
        for what in ("minimum", "maximum"):
            if processes > 1:
                # We create and destroy a new pool here in order to set the
                # objective direction for all reactions. This creates a
                # slight overhead but seems the most clean.
                chunk_size = len(reaction_ids) // processes
                pool = multiprocessing.Pool(
                    processes,
                    initializer=_init_worker,
                    initargs=(model, loopless, what[:3]),
                )
                for rxn_id, value in pool.imap_unordered(
                    _fva_step, reaction_ids, chunksize=chunk_size
                ):
                    fva_result.at[rxn_id, what] = value
                pool.close()
                pool.join()
            else:
                _init_worker(model, loopless, what[:3])
                for rxn_id, value in map(_fva_step, reaction_ids):
                    fva_result.at[rxn_id, what] = value

    return fva_result[["minimum", "maximum"]]


def find_blocked_reactions(
    model,
    reaction_list=None,
    zero_cutoff=None,
    open_exchanges=False,
    processes=None,
):
    """
    Find reactions that cannot carry any flux.

    The question whether or not a reaction is blocked is highly dependent
    on the current exchange reaction settings for a COBRA model. Hence an
    argument is provided to open all exchange reactions.

    Notes
    -----
    Sink and demand reactions are left untouched. Please modify them manually.

    Parameters
    ----------
    model : cobra.Model
        The model to analyze.
    reaction_list : list, optional
        List of reactions to consider, the default includes all model
        reactions.
    zero_cutoff : float, optional
        Flux value which is considered to effectively be zero
        (default model.tolerance).
    open_exchanges : bool, optional
        Whether or not to open all exchange reactions to very high flux ranges.
    processes : int, optional
        The number of parallel processes to run. Can speed up the computations
        if the number of reactions is large. If not explicitly
        passed, it will be set from the global configuration singleton.

    Returns
    -------
    list
        List with the identifiers of blocked reactions.

    """
    zero_cutoff = normalize_cutoff(model, zero_cutoff)

    with model:
        if open_exchanges:
            for reaction in model.exchanges:
                reaction.bounds = (
                    min(reaction.lower_bound, -1000),
                    max(reaction.upper_bound, 1000),
                )
        if reaction_list is None:
            reaction_list = model.reactions
        # Limit the search space to reactions which have zero flux. If the
        # reactions already carry flux in this solution,
        # then they cannot be blocked.
        model.slim_optimize()
        solution = get_solution(model, reactions=reaction_list)
        reaction_list = solution.fluxes[
            solution.fluxes.abs() < zero_cutoff
        ].index.tolist()
        # Run FVA to find reactions where both the minimal and maximal flux
        # are zero (below the cut off).
        flux_span = flux_variability_analysis(
            model,
            fraction_of_optimum=0.0,
            reaction_list=reaction_list,
            processes=processes,
        )
        return flux_span[
            flux_span.abs().max(axis=1) < zero_cutoff
        ].index.tolist()


def find_essential_genes(model, threshold=None, processes=None):
    """
    Return a set of essential genes.

    A gene is considered essential if restricting the flux of all reactions
    that depend on it to zero causes the objective, e.g., the growth rate,
    to also be zero, below the threshold, or infeasible.

    Parameters
    ----------
    model : cobra.Model
        The model to find the essential genes for.
    threshold : float, optional
        Minimal objective flux to be considered viable. By default this is
        1% of the maximal objective.
    processes : int, optional
        The number of parallel processes to run. If not passed,
        will be set to the number of CPUs found.
    processes : int, optional
        The number of parallel processes to run. Can speed up the computations
        if the number of knockouts to perform is large. If not explicitly
        passed, it will be set from the global configuration singleton.

    Returns
    -------
    set
        Set of essential genes
    """
    if threshold is None:
        threshold = model.slim_optimize(error_value=None) * 1e-02
    deletions = single_gene_deletion(model, method="fba", processes=processes)
    essential = deletions.loc[
        deletions["growth"].isna() | (deletions["growth"] < threshold), :
    ].index
    return {model.genes.get_by_id(g) for ids in essential for g in ids}


def find_essential_reactions(model, threshold=None, processes=None):
    """Return a set of essential reactions.

    A reaction is considered essential if restricting its flux to zero
    causes the objective, e.g., the growth rate, to also be zero, below the
    threshold, or infeasible.


    Parameters
    ----------
    model : cobra.Model
        The model to find the essential reactions for.
    threshold : float, optional
        Minimal objective flux to be considered viable. By default this is
        1% of the maximal objective.
    processes : int, optional
        The number of parallel processes to run. Can speed up the computations
        if the number of knockouts to perform is large. If not explicitly
        passed, it will be set from the global configuration singleton.

    Returns
    -------
    set
        Set of essential reactions
    """
    if threshold is None:
        threshold = model.slim_optimize(error_value=None) * 1e-02
    deletions = single_reaction_deletion(
        model, method="fba", processes=processes
    )
    essential = deletions.loc[
        deletions["growth"].isna() | (deletions["growth"] < threshold), :
    ].index
    return {model.reactions.get_by_id(r) for ids in essential for r in ids}



# -*- coding: utf-8 -*-

"""Define the MetaboliteSummary class."""

from __future__ import absolute_import, division

from operator import attrgetter

import pandas as pd

from cobra.core import get_solution
from cobra.core.summary import Summary
from cobra.flux_analysis.variability import flux_variability_analysis


class MetaboliteSummary(Summary):
    """
    Define the metabolite summary.

    Attributes
    ----------
    metabolite: cobra.Metabolite
        The metabolite to summarize.

    See Also
    --------
    Summary : Parent that defines further attributes.
    ReactionSummary
    ModelSummary

    """

    def __init__(self, metabolite, model, **kwargs):
        """
        Initialize a metabolite summary.

        Parameters
        ----------
        metabolite: cobra.Metabolite
            The metabolite object whose summary we intend to get.
        model : cobra.Model
            The metabolic model for which to generate a metabolite summary.

        Other Parameters
        ----------------
        kwargs :
            Further keyword arguments are passed on to the parent class.

        See Also
        --------
        Summary : Parent that has further default parameters.
        ReactionSummary
        ModelSummary

        """
        super(MetaboliteSummary, self).__init__(model=model, **kwargs)
        self.metabolite = metabolite

    def _generate(self):
        """
        Returns
        -------
        flux_summary: pandas.DataFrame
            The DataFrame of flux summary data.

        """
        if self.names:
            emit = attrgetter('name')
        else:
            emit = attrgetter('id')

        if self.solution is None:
            self.metabolite.model.slim_optimize(error_value=None)
            self.solution = get_solution(self.metabolite.model,
                                         reactions=self.metabolite.reactions)

        rxns = sorted(self.metabolite.reactions, key=attrgetter('id'))

        data = [
            (
                emit(rxn),
                self.solution[rxn.id] * rxn.metabolites[self.metabolite],
                rxn.build_reaction_string(use_metabolite_names=self.names),
                rxn
            ) for rxn in rxns
        ]

        flux_summary = pd.DataFrame.from_records(
            data=data,
            index=[rxn.id for rxn in rxns],
            columns=['id', 'flux', 'reaction_string', 'reaction']
        )

        assert flux_summary['flux'].sum() < self.metabolite.model.tolerance, \
            'Error in flux balance'

        if self.fva is not None:
            if hasattr(self.fva, 'columns'):
                fva_results = self.fva
            else:
                fva_results = flux_variability_analysis(
                    self.metabolite.model, list(self.metabolite.reactions),
                    fraction_of_optimum=self.fva)

            flux_summary = pd.concat([flux_summary, fva_results],
                                     axis=1, sort=False)
            flux_summary.rename(columns={'maximum': 'fmax', 'minimum': 'fmin'},
                                inplace=True)

            def set_min_and_max(row):
                """Scale and set proper min and max values for flux."""
                fmax = row.reaction.metabolites[self.metabolite] * row.fmax
                fmin = row.reaction.metabolites[self.metabolite] * row.fmin

                if abs(fmin) <= abs(fmax):
                    row.fmax = fmax
                    row.fmin = fmin
                else:
                    # Reverse fluxes
                    row.fmax = fmin
                    row.fmin = fmax

                return row

            flux_summary = flux_summary.apply(set_min_and_max, axis=1)

        flux_summary = self._process_flux_dataframe(flux_summary)

        total_flux = flux_summary.loc[flux_summary.is_input, 'flux'].sum()

        # Calculate flux percentage
        flux_summary['percent'] = (flux_summary['flux'] / total_flux) * 100

        return flux_summary

    def to_frame(self):
        """
        Returns
        -------
        A pandas.DataFrame of the summary.

        """
        flux_df = self._generate()

        flux_df['is_input'] = flux_df['is_input']\
            .apply(lambda x: 'PRODUCING' if x is True else 'CONSUMING')

        flux_df.columns = [col.upper() for col in flux_df.columns]

        if self.fva is not None:
            flux_df.rename(columns={'IS_INPUT': 'RXN_STAT',
                                    'FMIN': 'FLUX_MIN',
                                    'FMAX': 'FLUX_MAX'}, inplace=True)
            flux_df = flux_df[['RXN_STAT', 'ID', 'PERCENT', 'FLUX', 'FLUX_MIN',
                               'FLUX_MAX', 'REACTION_STRING']]
        else:
            flux_df.rename(columns={'IS_INPUT': 'RXN_STAT'}, inplace=True)
            flux_df = flux_df[['RXN_STAT', 'ID', 'PERCENT', 'FLUX',
                               'REACTION_STRING']]

        flux_df.set_index(['RXN_STAT', 'ID'], inplace=True)

        return flux_df

    def _to_table(self):
        """
        Returns
        -------
        A string of the summary table.

        """
        return self.to_frame().to_string(header=True, index=True, na_rep='',
                                         float_format=self.float_format,
                                         sparsify=True, justify='center')



# -*- coding: utf-8 -*-

"""Provide ACHR sampler."""

from __future__ import absolute_import, division

import numpy as np
import pandas

from cobra.sampling.hr_sampler import HRSampler, step


class ACHRSampler(HRSampler):
    """Artificial Centering Hit-and-Run sampler.

    A sampler with low memory footprint and good convergence.

    Parameters
    ----------
    model : cobra.Model
        The cobra model from which to generate samples.
    thinning : int, optional
        The thinning factor of the generated sampling chain. A thinning of 10
        means samples are returned every 10 steps.
    nproj : int > 0, optional
        How often to reproject the sampling point into the feasibility space.
        Avoids numerical issues at the cost of lower sampling. If you observe
        many equality constraint violations with `sampler.validate` you should
        lower this number.
    seed : int > 0, optional
        Sets the random number seed. Initialized to the current time stamp if
        None.

    Attributes
    ----------
    model : cobra.Model
        The cobra model from which the samples get generated.
    thinning : int
        The currently used thinning factor.
    n_samples : int
        The total number of samples that have been generated by this
        sampler instance.
    problem : collections.namedtuple
        A python object whose attributes define the entire sampling problem in
        matrix form. See docstring of `Problem`.
    warmup : numpy.matrix
        A matrix of with as many columns as reactions in the model and more
        than 3 rows containing a warmup sample in each row. None if no warmup
        points have been generated yet.
    retries : int
        The overall of sampling retries the sampler has observed. Larger
        values indicate numerical instabilities.
    seed : int > 0, optional
        Sets the random number seed. Initialized to the current time stamp if
        None.
    nproj : int
        How often to reproject the sampling point into the feasibility space.
    fwd_idx : numpy.array
        Has one entry for each reaction in the model containing the index of
        the respective forward variable.
    rev_idx : numpy.array
        Has one entry for each reaction in the model containing the index of
        the respective reverse variable.
    prev : numpy.array
        The current/last flux sample generated.
    center : numpy.array
        The center of the sampling space as estimated by the mean of all
        previously generated samples.

    Notes
    -----
    ACHR generates samples by choosing new directions from the sampling space's
    center and the warmup points. The implementation used here is the same
    as in the Matlab Cobra Toolbox [2]_ and uses only the initial warmup points
    to generate new directions and not any other previous iterates. This
    usually gives better mixing since the startup points are chosen to span
    the space in a wide manner. This also makes the generated sampling chain
    quasi-markovian since the center converges rapidly.

    Memory usage is roughly in the order of (2 * number reactions)^2
    due to the required nullspace matrices and warmup points. So large
    models easily take up a few GB of RAM.

    References
    ----------
    .. [1] Direction Choice for Accelerated Convergence in Hit-and-Run Sampling
       David E. Kaufman Robert L. Smith
       Operations Research 199846:1 , 84-95
       https://doi.org/10.1287/opre.46.1.84
    .. [2] https://github.com/opencobra/cobratoolbox

    """

    def __init__(self, model, thinning=100, nproj=None, seed=None):
        """Initialize a new ACHRSampler."""
        super(ACHRSampler, self).__init__(
            model, thinning, nproj=nproj, seed=seed
        )
        self.generate_fva_warmup()
        self.prev = self.center = self.warmup.mean(axis=0)
        np.random.seed(self._seed)

    def __single_iteration(self):
        pi = np.random.randint(self.n_warmup)

        # mix in the original warmup points to not get stuck
        delta = self.warmup[pi, :] - self.center
        self.prev = step(self, self.prev, delta)

        if self.problem.homogeneous and (
            self.n_samples * self.thinning % self.nproj == 0
        ):
            self.prev = self._reproject(self.prev)
            self.center = self._reproject(self.center)

        self.center = (self.n_samples * self.center) / (
            self.n_samples + 1
        ) + self.prev / (self.n_samples + 1)
        self.n_samples += 1

    def sample(self, n, fluxes=True):
        """Generate a set of samples.

        This is the basic sampling function for all hit-and-run samplers.

        Parameters
        ----------
        n : int
            The number of samples that are generated at once.
        fluxes : boolean
            Whether to return fluxes or the internal solver variables. If set
            to False will return a variable for each forward and backward flux
            as well as all additional variables you might have defined in the
            model.

        Returns
        -------
        numpy.matrix
            Returns a matrix with `n` rows, each containing a flux sample.

        Notes
        -----
        Performance of this function linearly depends on the number
        of reactions in your model and the thinning factor.

        """
        samples = np.zeros((n, self.warmup.shape[1]))

        for i in range(1, self.thinning * n + 1):
            self.__single_iteration()

            if i % self.thinning == 0:
                samples[i // self.thinning - 1, :] = self.prev

        if fluxes:
            names = [r.id for r in self.model.reactions]

            return pandas.DataFrame(
                samples[:, self.fwd_idx] - samples[:, self.rev_idx],
                columns=names,
            )
        else:
            names = [v.name for v in self.model.variables]

            return pandas.DataFrame(samples, columns=names)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import pandas as pd
import pytest

import cobra.medium as medium
from cobra import Metabolite, Reaction


class TestModelMedium:
    def test_model_medium(self, model):
        # Add a dummy 'malformed' import reaction
        bad_import = Reaction('bad_import')
        bad_import.add_metabolites({model.metabolites.pyr_c: 1})
        bad_import.bounds = (0, 42)
        model.add_reaction(bad_import)

        # Test basic setting and getting methods
        medium = model.medium
        model.medium = medium
        assert model.medium == medium

        # Test context management
        with model:
            # Ensure the bounds are correct beforehand
            assert model.reactions.EX_glc__D_e.lower_bound == -10
            assert model.reactions.bad_import.upper_bound == 42
            assert model.reactions.EX_co2_e.lower_bound == -1000

            # Make changes to the media
            new_medium = model.medium
            new_medium['EX_glc__D_e'] = 20
            new_medium['bad_import'] = 24
            del new_medium['EX_co2_e']

            # Change the medium, make sure changes work
            model.medium = new_medium
            assert model.reactions.EX_glc__D_e.lower_bound == -20
            assert model.reactions.bad_import.upper_bound == 24
            assert model.reactions.EX_co2_e.lower_bound == 0

        # Make sure changes revert after the contex
        assert model.reactions.EX_glc__D_e.lower_bound == -10
        assert model.reactions.bad_import.upper_bound == 42
        assert model.reactions.EX_co2_e.lower_bound == -1000

        new_medium['bogus_rxn'] = 0
        with pytest.raises(KeyError):
            model.medium = new_medium


class TestTypeDetection:

    def test_external_compartment(self, model):
        # by name
        assert medium.find_external_compartment(model) == "e"
        # from boundary counts
        for m in model.metabolites:
            if m.compartment == "e":
                m.compartment = "outside"
        for r in model.reactions:
            r._compartments = None
        assert medium.find_external_compartment(model) == "outside"
        # names are always right
        model.exchanges[0].reactants[0].compartment = "extracellular"
        assert medium.find_external_compartment(model) == "extracellular"

    def test_multi_external(self, model):
        for r in model.reactions:
            r._compartments = None
        model.exchanges[0].reactants[0].compartment = "extracellular"
        # still works due to different boundary numbers
        assert medium.find_external_compartment(model) == "e"
        model.exchanges[1].reactants[0].compartment = "extra cellular"
        model.remove_reactions(model.exchanges)
        # Now fails because same boundary count
        with pytest.raises(RuntimeError):
            medium.find_external_compartment(model)

    def test_exchange(self, model):
        ex = model.exchanges
        assert all(r.id.startswith("EX_") for r in ex)
        ex = medium.find_boundary_types(model, "exchange", "e")
        assert all(r.id.startswith("EX_") for r in ex)

    def test_demand(self, model):
        dm = Reaction("demand")
        model.add_reaction(dm)
        dm.build_reaction_from_string("atp_c ->")
        dm = model.demands
        assert len(dm) == 1
        assert "demand" in [r.id for r in dm]

    def test_sink(self, model):
        sn = Reaction("sink")
        model.add_reaction(sn)
        sn.build_reaction_from_string("atp_c <->")
        sn.bounds = -1000, 1000
        sn = model.sinks
        assert len(sn) == 1
        assert "sink" in [r.id for r in sn]

    def test_sbo_terms(self, model):
        assert not medium.is_boundary_type(
            model.reactions.ATPM, "exchange", "e")
        model.reactions.ATPM.annotation["sbo"] = "SBO:0000627"
        assert medium.is_boundary_type(model.reactions.ATPM, "exchange", "bla")
        model.reactions.ATPM.annotation["sbo"] = "SBO:0000632"
        assert not medium.is_boundary_type(
            model.reactions.ATPM, "exchange", "e")


class TestMinimalMedia:

    def test_medium_linear(self, model):
        med = medium.minimal_medium(model, 0.8)
        assert len(med) <= 4
        assert all(med > 1e-6)

    def test_medium_mip(self, model):
        med = medium.minimal_medium(model, 0.8, minimize_components=True)
        assert len(med) <= 4
        assert all(med > 1e-6)

        # Anaerobic growth
        med = medium.minimal_medium(model, 0.1, minimize_components=True)
        assert len(med) <= 3
        assert all(med > 1e-6)

    def test_medium_alternative_mip(self, model):
        med = medium.minimal_medium(model, 0.8, minimize_components=5,
                                    open_exchanges=True)
        assert isinstance(med, pd.DataFrame)
        assert med.shape[0] >= 5
        assert med.shape[1] == 5
        assert all((med > 0).sum() == 3)
        assert all(med.sum(axis=1) > 1e-6)

    def test_benchmark_medium_linear(self, model, benchmark):
        benchmark(medium.minimal_medium, model, 0.8)

    def test_benchmark_medium_mip(self, model, benchmark):
        benchmark(medium.minimal_medium, model, 0.8, True)

    def test_medium_exports(self, model):
        med = medium.minimal_medium(model, 0.8, exports=True,
                                    minimize_components=True)
        assert len(med) > 4
        assert any(med < -1e-6)

    def test_open_exchanges(self, model):
        model.reactions.EX_glc__D_e.bounds = 0, 0
        med = medium.minimal_medium(model, 0.8)
        assert med is None
        med = medium.minimal_medium(model, 0.8, minimize_components=True)
        assert med is None

        med = medium.minimal_medium(model, 0.8, open_exchanges=True)
        assert len(med) >= 3
        med = medium.minimal_medium(model, 0.8, open_exchanges=100)
        assert len(med) >= 3


class TestErrorsAndExceptions:

    def test_no_boundary_reactions(self, empty_model):
        assert medium.find_boundary_types(empty_model, 'e', None) == []

    def test_no_names_or_boundary_reactions(self, empty_model):
        with pytest.raises(RuntimeError):
            medium.find_external_compartment(empty_model)

    def test_bad_exchange(self, model):
        with pytest.raises(ValueError):
            m = Metabolite("baddy", compartment="nonsense")
            model.add_boundary(m, type="exchange")
        m = Metabolite("goody", compartment="e")
        rxn = model.add_boundary(m, type="exchange")
        assert isinstance(rxn, Reaction)



# -*- coding: utf-8 -*-

"""Test functionalities provided by mat.py"""

from __future__ import absolute_import

from os.path import join
from pickle import load

import pytest

from cobra import io
from cobra.test.test_io.conftest import compare_models


try:
    import scipy
except ImportError:
    scipy = None


@pytest.fixture(scope="function")
def raven_model(data_directory):
    """Fixture for RAVEN model."""
    with open(join(data_directory, "raven.pickle"), "rb") as infile:
        return load(infile)


@pytest.mark.skipif(scipy is None, reason='scipy unavailable')
# @pytest.mark.parametrize("ref_model, filename",
#                          [(pytest.fixture_request("mini_model"),
#                            "mini.mat"),
#                           (pytest.fixture_request("raven_model"),
#                            "raven.mat")])
# TODO: wait for pytest.fixture_request() to get approved
def test_load_matlab_model(data_directory, mini_model, raven_model):
    """Test the reading of MAT model."""
    mini_mat_model = io.load_matlab_model(join(data_directory, "mini.mat"))
    raven_mat_model = io.load_matlab_model(join(data_directory, "raven.mat"))
    assert compare_models(mini_model, mini_mat_model) is None
    assert compare_models(raven_model, raven_mat_model) is None


# @pytest.mark.xfail(reason="localPath not supported yet")
@pytest.mark.skipif(scipy is None, reason='scipy unavailable')
# @pytest.mark.parametrize("model, filename",
#                          [(pytest.fixture_request("mini_model"),
#                            "mini.mat"),
#                           (pytest.fixture_request("raven_model"),
#                            "raven.mat")])
# TODO: wait for pytest.fixture_request() to get approved
def test_save_matlab_model(tmpdir, mini_model, raven_model):
    """Test the writing of MAT model."""
    mini_output_file = tmpdir.join("mini.mat")
    raven_output_file = tmpdir.join("raven.mat")
    # scipy.io.savemat() doesn't support anything other than
    # str or file-stream object, hence the str conversion
    io.save_matlab_model(mini_model, str(mini_output_file))
    io.save_matlab_model(raven_model, str(raven_output_file))
    assert mini_output_file.check()
    assert raven_output_file.check()



# -*- coding: utf-8 -*-
"""
Testing SBML functionality based on libsbml.
"""

from __future__ import absolute_import

from collections import namedtuple
from os import unlink
from os.path import join, split
from pickle import load
from tempfile import gettempdir

import pytest

import cobra
from cobra import Model
from cobra.io import read_sbml_model, validate_sbml_model, write_sbml_model


config = cobra.Configuration()  # for default bounds

try:
    import jsonschema
except ImportError:
    jsonschema = None

# ----------------------------------
# Definition of SBML files to test
# ----------------------------------
IOTrial = namedtuple('IOTrial',
                     ['name', 'reference_file', 'test_file', 'read_function',
                      'write_function', 'validation_function'])
trials = [IOTrial('fbc2', 'mini.pickle', 'mini_fbc2.xml',
                  read_sbml_model, write_sbml_model,
                  validate_sbml_model),
          IOTrial('fbc2Gz', 'mini.pickle', 'mini_fbc2.xml.gz',
                  read_sbml_model, write_sbml_model, None),
          IOTrial('fbc2Bz2', 'mini.pickle', 'mini_fbc2.xml.bz2',
                  read_sbml_model, write_sbml_model, None),
          IOTrial('fbc1', 'mini.pickle', 'mini_fbc1.xml',
                  read_sbml_model, write_sbml_model, None),
          IOTrial('cobra', None, 'mini_cobra.xml',
                  read_sbml_model, write_sbml_model, None),
          ]
trial_names = [node.name for node in trials]


@pytest.mark.parametrize("trial", trials)
def test_validate(trial, data_directory):
    """ Test validation function. """
    if trial.validation_function is None:
        pytest.skip('not implemented')
    test_file = join(data_directory, trial.test_file)
    trial.validation_function(test_file)


class TestCobraIO:
    """ Tests the read and write functions. """

    @classmethod
    def compare_models(cls, name, model1, model2):
        assert len(model1.reactions) == len(model2.reactions)
        assert len(model1.metabolites) == len(model2.metabolites)
        assert model1.objective.direction == model2.objective.direction
        for attr in ("id", "name", "lower_bound", "upper_bound",
                     "objective_coefficient", "gene_reaction_rule"):
            assert getattr(model1.reactions[0], attr) == getattr(
                model2.reactions[0], attr)
            assert getattr(model1.reactions[5], attr) == getattr(
                model2.reactions[5], attr)
            assert getattr(model1.reactions[-1], attr) == getattr(
                model2.reactions[-1], attr)
        for attr in ("id", "name", "compartment", "formula", "charge"):
            assert getattr(model1.metabolites[0], attr) == getattr(
                model2.metabolites[0], attr)
            assert getattr(model1.metabolites[5], attr) == getattr(
                model2.metabolites[5], attr)
            assert getattr(model1.metabolites[-1], attr) == getattr(
                model2.metabolites[-1], attr)
        assert len(model1.reactions[0].metabolites) == len(
            model2.reactions[0].metabolites)
        assert len(model1.reactions[8].metabolites) == len(
            model2.reactions[8].metabolites)
        assert len(model1.reactions[-1].metabolites) == len(
            model2.reactions[-1].metabolites)
        assert len(model1.genes) == len(model2.genes)

        # ensure they have the same solution max
        solution1 = model1.optimize()
        solution2 = model2.optimize()
        assert (solution1.status == 'infeasible' and
                solution2.status == 'infeasible') or \
            abs(solution1.objective_value -
                solution2.objective_value) < 0.001
        # ensure the references are correct
        assert model2.metabolites[0]._model is model2
        assert model2.reactions[0]._model is model2
        assert model2.genes[0]._model is model2

    @classmethod
    def extra_comparisons(cls, name, model1, model2):
        assert model1.compartments == model2.compartments

        # FIXME: problems of duplicate annotations in test data
        #  ('cas': ['56-65-5', '56-65-5'])
        # assert dict(model1.metabolites[4].annotation) == dict(
        #    model2.metabolites[4].annotation)
        d1 = model1.reactions[4].annotation
        d2 = model2.reactions[4].annotation
        assert list(d1.keys()) == list(d2.keys())
        for k in d1:
            assert set(d1[k]) == set(d2[k])
        assert dict(model1.reactions[4].annotation) == dict(
            model2.reactions[4].annotation)
        assert dict(model1.genes[5].annotation) == dict(
            model2.genes[5].annotation)

        for attr in ("id", "name"):
            assert getattr(model1.genes[0], attr) == getattr(model2.genes[0],
                                                             attr)
            assert getattr(model1.genes[10], attr) == getattr(model2.genes[10],
                                                              attr)
            assert getattr(model1.genes[-1], attr) == getattr(model2.genes[-1],
                                                              attr)

    def test_read_1(self, io_trial):
        name, reference_model, test_model, _ = io_trial
        if name in ['fbc1']:
            pytest.xfail('not supported')
        if reference_model:
            self.compare_models(name, reference_model, test_model)

    def test_read_2(self, io_trial):
        name, reference_model, test_model, _ = io_trial
        if name in ['fbc1', 'mat', 'cobra', 'raven-mat']:
            pytest.xfail('not supported')
        if reference_model:
            self.extra_comparisons(name, reference_model, test_model)

    def test_write_1(self, io_trial):
        name, _, test_model, reread_model = io_trial
        if name in ['fbc1', 'raven-mat']:
            pytest.xfail('not supported')

        self.compare_models(name, test_model, reread_model)

    def test_write_2(self, io_trial):
        name, _, test_model, reread_model = io_trial
        if name in ['fbc1', 'mat', 'cobra', 'raven-mat']:
            pytest.xfail('not supported')
        self.extra_comparisons(name, test_model, reread_model)


@pytest.fixture(scope="module", params=trials, ids=trial_names)
def io_trial(request, data_directory):
    reference_model = None
    if request.param.reference_file:
        with open(join(data_directory, request.param.reference_file),
                  "rb") as infile:
            reference_model = load(infile)
    test_model = request.param.read_function(join(data_directory,
                                                  request.param.test_file))
    test_output_filename = join(gettempdir(),
                                split(request.param.test_file)[-1])
    # test writing the model within a context with a non-empty stack
    with test_model:
        test_model.objective = test_model.objective
        request.param.write_function(test_model, test_output_filename)
    reread_model = request.param.read_function(test_output_filename)
    unlink(test_output_filename)
    return request.param.name, reference_model, test_model, reread_model


def test_filehandle(data_directory, tmp_path):
    """Test reading and writing to file handle."""
    with open(join(data_directory, "mini_fbc2.xml"), "r") as f_in:
        model1 = read_sbml_model(f_in)
        assert model1 is not None

    sbml_path = join(str(tmp_path), "test.xml")
    with open(sbml_path, "w") as f_out:
        write_sbml_model(model1, f_out)

    with open(sbml_path, "r") as f_in:
        model2 = read_sbml_model(f_in)

    TestCobraIO.compare_models(name="filehandle",
                               model1=model1, model2=model2)


def test_from_sbml_string(data_directory):
    """Test reading from SBML string."""
    sbml_path = join(data_directory, "mini_fbc2.xml")
    with open(sbml_path, "r") as f_in:
        sbml_str = f_in.read()
        model1 = read_sbml_model(sbml_str)

    model2 = read_sbml_model(sbml_path)
    TestCobraIO.compare_models(name="read from string",
                               model1=model1, model2=model2)


@pytest.mark.skip(reason="Model history currently not written")
def test_model_history(tmp_path):
    """Testing reading and writing of ModelHistory."""
    model = Model("test")
    model._sbml = {
        "creators": [{
            "familyName": "Mustermann",
            "givenName": "Max",
            "organisation": "Muster University",
            "email": "muster@university.com",
        }]
    }

    sbml_path = join(str(tmp_path), "test.xml")
    with open(sbml_path, "w") as f_out:
        write_sbml_model(model, f_out)

    with open(sbml_path, "r") as f_in:
        model2 = read_sbml_model(f_in)

    assert "creators" in model2._sbml
    assert len(model2._sbml["creators"]) is 1
    c = model2._sbml["creators"][0]
    assert c["familyName"] == "Mustermann"
    assert c["givenName"] == "Max"
    assert c["organisation"] == "Muster University"
    assert c["email"] == "muster@university.com"


def test_groups(data_directory, tmp_path):
    """Testing reading and writing of groups"""
    sbml_path = join(data_directory, "e_coli_core.xml")
    model = read_sbml_model(sbml_path)
    assert model.groups is not None
    assert len(model.groups) == 10
    g1 = model.groups[0]
    assert len(g1.members) == 6

    temp_path = join(str(tmp_path), "test.xml")
    with open(temp_path, "w") as f_out:
        write_sbml_model(model, f_out)

    with open(temp_path, "r") as f_in:
        model2 = read_sbml_model(f_in)

        assert model2.groups is not None
        assert len(model2.groups) == 10
        g1 = model2.groups[0]
        assert len(g1.members) == 6


def test_missing_flux_bounds1(data_directory):
    sbml_path = join(data_directory, "annotation.xml")
    with open(sbml_path, "r") as f_in:
        # missing flux bounds are set to cobra.configuration.bounds
        model, errors = validate_sbml_model(f_in,
                                            set_missing_bounds=True)
        r1 = model.reactions.R1
        assert r1.lower_bound == config.lower_bound
        assert r1.upper_bound == config.upper_bound


def test_missing_flux_bounds2(data_directory):
    sbml_path = join(data_directory, "annotation.xml")
    with open(sbml_path, "r") as f_in:
        # missing flux bounds are set to [-INF, INF]
        model, errors = validate_sbml_model(f_in,
                                            set_missing_bounds=False)
        r1 = model.reactions.R1
        assert r1.lower_bound == config.lower_bound
        assert r1.upper_bound == config.upper_bound


def test_validate(data_directory):
    """Test the validation code. """
    sbml_path = join(data_directory, "mini_fbc2.xml")
    with open(sbml_path, "r") as f_in:
        model1, errors = validate_sbml_model(f_in,
                                             check_modeling_practice=True)
        assert model1
        assert errors
        assert len(errors["SBML_WARNING"]) == 0


def test_validation_warnings(data_directory):
    """Test the validation warnings. """
    sbml_path = join(data_directory, "validation.xml")
    with open(sbml_path, "r") as f_in:
        model1, errors = validate_sbml_model(f_in,
                                             check_modeling_practice=True)
        assert model1
        assert errors
        assert len(errors["COBRA_WARNING"]) == 3
        assert "No objective in listOfObjectives" in errors["COBRA_WARNING"]


def test_infinity_bounds(data_directory, tmp_path):
    """Test infinity bound example. """
    sbml_path = join(data_directory, "fbc_ex1.xml")
    model = read_sbml_model(sbml_path)

    # check that simulation works
    solution = model.optimize()

    # check that values are set
    r = model.reactions.get_by_id("EX_X")
    assert r.lower_bound == -float("Inf")
    assert r.upper_bound == float("Inf")

    temp_path = join(str(tmp_path), "test.xml")
    with open(temp_path, "w") as f_out:
        write_sbml_model(model, f_out)

    with open(temp_path, "r") as f_in:
        model2 = read_sbml_model(f_in)
        r = model2.reactions.get_by_id("EX_X")
        assert r.lower_bound == -float("Inf")
        assert r.upper_bound == float("Inf")


def test_boundary_conditions(data_directory):
    """Test infinity bound example. """
    sbml_path1 = join(data_directory, "fbc_ex1.xml")
    model1 = read_sbml_model(sbml_path1)
    sol1 = model1.optimize()

    # model with species boundaryCondition==True
    sbml_path2 = join(data_directory, "fbc_ex2.xml")
    model2 = read_sbml_model(sbml_path2)
    sol2 = model2.optimize()

    r = model2.reactions.get_by_id("EX_X")
    assert r.lower_bound == config.lower_bound
    assert r.upper_bound == config.upper_bound

    assert sol1.objective_value == sol2.objective_value


def test_gprs(data_directory, tmp_path):
    """Test that GPRs are written and read correctly"""
    model1 = read_sbml_model(join(data_directory, "iJO1366.xml.gz"))

    sbml_path = join(str(tmp_path), "test.xml")
    with open(sbml_path, "w") as f_out:
        write_sbml_model(model1, f_out)

    with open(sbml_path, "r") as f_in:
        model2 = read_sbml_model(f_in)

    for r1 in model1.reactions:
        rid = r1.id
        r2 = model2.reactions.get_by_id(rid)
        gpr1 = r1.gene_reaction_rule
        gpr2 = r2.gene_reaction_rule

        assert gpr1 == gpr2


def test_identifiers_annotation():
    from cobra.io.sbml import _parse_annotation_info

    for uri in [
        "http://identifiers.org/chebi/CHEBI:000123",
        "https://identifiers.org/chebi/CHEBI:000123",
        "http://identifiers.org/CHEBI:000123",
        "https://identifiers.org/CHEBI:000123",
    ]:
        data = _parse_annotation_info(uri)
        assert data
        assert data[0] == "chebi"
        assert data[1] == "CHEBI:000123"

    for uri in [
        "http://identifiers.org/taxonomy/9602",
        "https://identifiers.org/taxonomy/9602",
        "http://identifiers.org/taxonomy:9602",
        "https://identifiers.org/taxonomy:9602",
    ]:
        data = _parse_annotation_info(uri)
        assert data
        assert data[0] == "taxonomy"
        assert data[1] == "9602"

    for uri in [
        "http://identifier.org/taxonomy/9602",
        "https://test.com",
    ]:
        data = _parse_annotation_info(uri)
        assert data is None


def test_smbl_with_notes(data_directory, tmp_path):
    """Test that NOTES in the RECON 2.2 style are written and read correctly"""
    sbml_path = join(data_directory, "example_notes.xml")
    model = read_sbml_model(sbml_path)
    assert model.metabolites is not None
    metabolite_notes = {
        '2hb_e': {'CHARGE': '-1', 'FORMULA': 'C4H7O3',
                  'SMILES': 'CCC(O)C(O)=O'},
        'nad_e': {'CHARGE': '-1', 'FORMULA': 'C21H26N7O14P2',
                  'SMILES': 'NC(=O)c1ccc[n+](c1)[C@@H]1O[C@H](COP([O-])('
                            '=O)OP([O-])(=O)OC[C@H]2O[C@H]([C@H](O)['
                            'C@@H]2O)n2cnc3c(N)ncnc23)[C@@H](O)[C@H]1O'},
        'h_e': {'CHARGE': '1', 'FORMULA': 'H', 'SMILES': '[1H+]'},
        '2obut_e': {'CHARGE': '-1', 'FORMULA': 'C4H5O3',
                    'SMILES': 'CCC(=O)C([O-])=O'},
        'nadh_e': {'CHARGE': '-2', 'FORMULA': 'C21H27N7O14P2',
                   'SMILES': 'NC(=O)C1=CN(C=CC1)[C@@H]1O[C@H](COP([O-])('
                             '=O)OP([O-])(=O)OC[C@H]2O[C@H]([C@H](O)['
                             'C@@H]2O)n2cnc3c(N)ncnc23)[C@@H](O)[C@H]1O'}
    }
    metabolite_annotations = {
        '2hb_e': {'sbo': 'SBO:0000247',
                  'inchi': 'InChI=1S/C4H8O3/c1-2-3(5)4(6)7/h3,5H,2H2,1H3,'
                           '(H,6,7)',
                  'chebi': 'CHEBI:1148'},
        'nad_e': {'sbo': 'SBO:0000247',
                  'inchi': 'InChI=1S/C21H27N7O14P2/c22-17-12-19('
                           '25-7-24-17)28(8-26-12)21-16(32)14(30)11('
                           '41-21)6-39-44(36,37)42-43(34,35)38-5-10-13(29)15('
                           '31)20(40-10)27-3-1-2-9(4-27)18('
                           '23)33/h1-4,7-8,10-11,13-16,20-21,29-32H,5-6H2,'
                           '(H5-,22,23,24,25,33,34,35,36,37)/p-1/t10-,'
                           '11-,13-,14-,15-,16-,20-,21-/m1/s1',
                  'chebi': 'CHEBI:57540'},
        'h_e': {'sbo': 'SBO:0000247', 'inchi': 'InChI=1S/p+1/i/hH',
                'chebi': 'CHEBI:24636'},
        '2obut_e': {'sbo': 'SBO:0000247',
                    'inchi': 'InChI=1S/C4H6O3/c1-2-3(5)4(6)7/h2H2,1H3,(H,6,'
                             '7)/p-1',
                    'chebi': 'CHEBI:16763'},
        'nadh_e': {'sbo': 'SBO:0000247',
                   'inchi': 'InChI=1S/C21H29N7O14P2/c22-17-12-19('
                            '25-7-24-17)28(8-26-12)21-16(32)14(30)11('
                            '41-21)6-39-44(36,37)42-43(34,35)38-5-10-13('
                            '29)15(31)20(40-10)27-3-1-2-9(4-27)18('
                            '23)33/h1,3-4,7-8,10-11,13-16,20-21,29-32H,2,'
                            '5-6H2,(H2,23,33)(H,34,35)(H,36,37)(H2,22,24,'
                            '25)/p-2/t10-,11-,13-,14-,15-,16-,20-,21-/m1/s1',
                   'chebi': 'CHEBI:57945'}
    }
    reaction_notes = {'CONFIDENCE_LEVEL': '4', 'NOTES': 'NCD',
                      'SUBSYSTEM': 'Propanoate metabolism',
                      'GENE_ASSOCIATION': '(HGNC:8546 and HGNC:8548) or'
                                          ' (HGNC:8547 and HGNC:8548)'}
    reaction_annotations = {'sbo': 'SBO:0000176', 'ec-code': '1.1.1.27',
                            'pubmed': ['10108', '21765']}

    for met_id in metabolite_notes:
        assert model.metabolites.has_id(met_id)
        for note_key in metabolite_notes[met_id].keys():
            assert note_key in model.metabolites.get_by_id(met_id).notes
            assert metabolite_notes[met_id][note_key] == \
                model.metabolites.get_by_id(met_id).notes[note_key]
        for annotation_key in metabolite_annotations[met_id].keys():
            assert annotation_key in model.metabolites.get_by_id(
                met_id).annotation
            print(met_id)
            assert metabolite_annotations[met_id][annotation_key] == \
                model.metabolites.get_by_id(
                    met_id).annotation[annotation_key]

    for note_key in reaction_notes.keys():
        assert note_key in model.reactions[0].notes.keys()
        assert reaction_notes[note_key] == model.reactions[0].notes[note_key]
    assert model.reactions[
               0].gene_reaction_rule == '(HGNC:8546 and HGNC:8548) or ' \
                                        '(HGNC:8547 and HGNC:8548)'
    assert len(model.groups) == 1
    for annotation_key in reaction_annotations.keys():
        assert annotation_key in model.reactions[0].annotation.keys()
        assert reaction_annotations[annotation_key] == model.reactions[
            0].annotation[annotation_key]



# -*- coding: utf-8 -*-

from __future__ import absolute_import


NOT_MASS_BALANCED_TERMS = {"SBO:0000627",  # EXCHANGE
                           "SBO:0000628",  # DEMAND
                           "SBO:0000629",  # BIOMASS
                           "SBO:0000631",  # PSEUDOREACTION
                           "SBO:0000632",  # SINK
                           }


def check_mass_balance(model):
    unbalanced = {}
    for reaction in model.reactions:
        if reaction.annotation.get("sbo") not in NOT_MASS_BALANCED_TERMS:
            balance = reaction.check_mass_balance()
            if balance:
                unbalanced[reaction] = balance
    return unbalanced


def check_metabolite_compartment_formula(model):
    errors = []
    for met in model.metabolites:
        if met.formula is not None and len(met.formula) > 0:
            if not met.formula.isalnum():
                errors.append("Metabolite '%s' formula '%s' not alphanumeric" %
                              (met.id, met.formula))
    return errors



# -*- coding: utf-8 -*-

"""Define the group class."""

from __future__ import absolute_import

from warnings import warn

from six import string_types

from cobra.core.dictlist import DictList
from cobra.core.object import Object


class Group(Object):
    """
    Manage groups via this implementation of the SBML group specification.

    `Group` is a class for holding information regarding a pathways,
    subsystems, or other custom groupings of objects within a cobra.Model
    object.

    Parameters
    ----------
    id : str
        The identifier to associate with this group
    name : str, optional
        A human readable name for the group
    members : iterable, optional
        A DictList containing references to cobra.Model-associated objects
        that belong to the group.
    kind : {"collection", "classification", "partonomy"}, optional
        The kind of group, as specified for the Groups feature in the SBML
        level 3 package specification. Can be any of "classification",
        "partonomy", or "collection". The default is "collection".
        Please consult the SBML level 3 package specification to ensure you
        are using the proper value for kind. In short, members of a
        "classification" group should have an "is-a" relationship to the group
        (e.g. member is-a polar compound, or member is-a transporter).
        Members of a "partonomy" group should have a "part-of" relationship
        (e.g. member is part-of glycolysis). Members of a "collection" group
        do not have an implied relationship between the members, so use this
        value for kind when in doubt (e.g. member is a gap-filled reaction,
        or member is involved in a disease phenotype).
    """
    KIND_TYPES = ("collection", "classification", "partonomy")

    def __init__(self, id, name='', members=None, kind=None):
        Object.__init__(self, id, name)

        self._members = DictList() if members is None else DictList(members)
        self._kind = None
        self.kind = "collection" if kind is None else kind
        # self.model is None or refers to the cobra.Model that
        # contains self
        self._model = None

    def __len__(self):
        return len(self._members)

    # read-only
    @property
    def members(self):
        return self._members

    @property
    def kind(self):
        return self._kind

    @kind.setter
    def kind(self, kind):
        kind = kind.lower()
        if kind in self.KIND_TYPES:
            self._kind = kind
        else:
            raise ValueError(
                "Kind can only by one of: {}.".format(", ".join(
                    self.KIND_TYPES)))

    def add_members(self, new_members):
        """
        Add objects to the group.

        Parameters
        ----------
        new_members : list
            A list of cobrapy objects to add to the group.

        """

        if isinstance(new_members, string_types) or \
                hasattr(new_members, "id"):
            warn("need to pass in a list")
            new_members = [new_members]

        self._members.union(new_members)

    def remove_members(self, to_remove):
        """
        Remove objects from the group.

        Parameters
        ----------
        to_remove : list
            A list of cobra objects to remove from the group
        """

        if isinstance(to_remove, string_types) or \
                hasattr(to_remove, "id"):
            warn("need to pass in a list")
            to_remove = [to_remove]

        for member_to_remove in to_remove:
            self._members.remove(member_to_remove)



# -*- coding: utf-8 -*-

"""Test data storage and recovery using pickle."""

from __future__ import absolute_import

from os.path import join
from pickle import dump, load

import pytest

from cobra.test.test_io.conftest import compare_models


try:
    import cPickle
    cload = cPickle.load
    cdump = cPickle.dump
except ImportError:
    cload = None
    cdump = None


@pytest.mark.parametrize("load_function", [load, cload])
def test_read_pickle(data_directory, mini_model, load_function):
    """Test the reading of model from pickle."""
    if load_function is None:
        pytest.skip()

    with open(join(data_directory, "mini.pickle"), "rb") as infile:
        pickle_model = load_function(infile)

    assert compare_models(mini_model, pickle_model) is None


@pytest.mark.parametrize("dump_function", [dump, cdump])
def test_write_pickle(tmpdir, mini_model, dump_function):
    """Test the writing of model to pickle."""
    if dump_function is None:
        pytest.skip()

    output_file = tmpdir.join("mini.pickle")
    with open(str(output_file), "wb") as outfile:
        dump_function(mini_model, outfile)

    assert output_file.check()



from os.path import join

import pytest

from cobra.io import load_json_model, write_sbml_model


def test_load_json_model_valid(data_directory, tmp_path):
    """Test loading a valid annotation from JSON."""
    path_to_file = join(data_directory, "valid_annotation_format.json")
    model = load_json_model(path_to_file)
    expected = {
        'bigg.reaction': [['is', 'PFK26']],
        'kegg.reaction': [['is', 'R02732']],
        'rhea': [['is', '15656']]
    }
    for metabolite in model.metabolites:
        assert metabolite.annotation == expected
    path_to_output = join(str(tmp_path), 'valid_annotation_output.xml')
    write_sbml_model(model, path_to_output)


def test_load_json_model_invalid(data_directory):
    """Test that loading an invalid annotation from JSON raises TypeError"""
    path = join(data_directory, "invalid_annotation_format.json")
    with pytest.raises(TypeError):
        model = load_json_model(path)



# -*- coding: utf-8 -*-

"""Define the ModelSummary class."""

from __future__ import absolute_import

import logging
from collections import OrderedDict
from operator import attrgetter

import numpy as np
import pandas as pd
from six import iteritems, iterkeys

from cobra.core import get_solution
from cobra.core.summary import Summary
from cobra.flux_analysis.variability import flux_variability_analysis
from cobra.util.solver import linear_reaction_coefficients


logger = logging.getLogger(__name__)


class ModelSummary(Summary):
    """
    Define the model summary.

    See Also
    --------
    Summary : Parent that defines further attributes.
    MetaboliteSummary
    ReactionSummary

    """

    def __init__(self, model, **kwargs):
        """
        Initialize a model summary.

        Parameters
        ----------
        model : cobra.Model
            The metabolic model for which to generate a summary.

        Other Parameters
        ----------------
        kwargs :
            Further keyword arguments are passed on to the parent class.

        See Also
        --------
        Summary : Parent that has further default parameters.
        MetaboliteSummary
        ReactionSummary

        """
        super(ModelSummary, self).__init__(model=model, **kwargs)

    def _generate(self):
        """
        Returns
        -------
        flux_summary: pandas.DataFrame
            The DataFrame of flux summary data.

        """
        if self.names:
            emit = attrgetter('name')
        else:
            emit = attrgetter('id')

        obj_rxns = linear_reaction_coefficients(self.model)
        boundary_rxns = self.model.exchanges
        summary_rxns = set(obj_rxns.keys()).union(boundary_rxns)

        if self.solution is None:
            self.model.slim_optimize(error_value=None)
            self.solution = get_solution(self.model, reactions=summary_rxns)

        # the order needs to be maintained
        ord_obj_rxns = OrderedDict(obj_rxns)

        obj_data = [(emit(rxn), self.solution[rxn.id] * stoich, np.nan, rxn,
                     np.nan, 1.0) for rxn, stoich in iteritems(ord_obj_rxns)]

        # create a DataFrame of objective reactions
        obj_df = pd.DataFrame.from_records(
            data=obj_data,
            index=[rxn.id for rxn in iterkeys(ord_obj_rxns)],
            columns=['id', 'flux', 'metabolite', 'reaction', 'is_input',
                     'is_obj_rxn']
        )

        # create a collection of metabolites from the boundary reactions
        mets = OrderedDict((met, rxn) for rxn in boundary_rxns
                           for met in sorted(rxn.metabolites,
                                             key=attrgetter('id')))
        index = [met.id for met in iterkeys(mets)]

        met_data = [(emit(met), self.solution[rxn.id] * rxn.metabolites[met],
                     met, rxn) for met, rxn in iteritems(mets)]

        # create a DataFrame of metabolites
        flux_summary = pd.DataFrame.from_records(
            data=met_data,
            index=index,
            columns=['id', 'flux', 'metabolite', 'reaction']
        )

        # Calculate FVA results if requested
        if self.fva is not None:
            if len(index) != len(boundary_rxns):
                logger.warning(
                    'There exists more than one boundary reaction per '
                    'metabolite. Please be careful when evaluating flux '
                    'ranges.')

            if hasattr(self.fva, 'columns'):
                fva_results = self.fva
            else:
                fva_results = flux_variability_analysis(
                    self.model, reaction_list=boundary_rxns,
                    fraction_of_optimum=self.fva
                )

            # save old index
            old_index = flux_summary.index
            # generate new index for consistency with fva_results
            flux_summary['rxn_id'] = flux_summary.apply(
                lambda x: x.reaction.id, axis=1
            )
            # change to new index
            flux_summary.set_index(['rxn_id'], inplace=True)
            flux_summary = pd.concat([flux_summary, fva_results],
                                     axis=1, sort=False)
            flux_summary.rename(columns={'maximum': 'fmax', 'minimum': 'fmin'},
                                inplace=True)
            # revert to old index
            flux_summary.set_index(old_index)

            def set_min_and_max(row):
                """Scale and set proper min and max values for flux."""
                fmax = row.reaction.metabolites[row.metabolite] * row.fmax
                fmin = row.reaction.metabolites[row.metabolite] * row.fmin

                if abs(fmin) <= abs(fmax):
                    row.fmax = fmax
                    row.fmin = fmin
                else:
                    # Reverse fluxes
                    row.fmax = fmin
                    row.fmin = fmax

                return row

            flux_summary = flux_summary.apply(set_min_and_max, axis=1)

        flux_summary = self._process_flux_dataframe(flux_summary)

        flux_summary = pd.concat([flux_summary, obj_df], sort=False)

        return flux_summary

    def to_frame(self):
        """
        Returns
        -------
        A pandas.DataFrame of the summary.

        """
        flux_df = self._generate()

        # obtain separate DataFrames and join them instead of reorganizing the
        # old DataFrame, since that is easier to understand

        if self.fva is not None:
            column_names = ['IN_FLUXES-ID', 'IN_FLUXES-FLUX',
                            'IN_FLUXES-FLUX_MIN', 'IN_FLUXES-FLUX_MAX',
                            'OUT_FLUXES-ID', 'OUT_FLUXES-FLUX',
                            'OUT_FLUXES-FLUX_MIN', 'OUT_FLUXES-FLUX_MAX',
                            'OBJECTIVES-ID', 'OBJECTIVES-FLUX']

            # generate separate DataFrames
            met_in_df = flux_df[flux_df.is_input == 1.0]\
                .loc[:, ['id', 'flux', 'fmin', 'fmax']].reset_index(drop=True)

            met_out_df = flux_df[flux_df.is_input == 0.0]\
                .loc[:, ['id', 'flux', 'fmin', 'fmax']].reset_index(drop=True)

            obj_df = flux_df[flux_df.is_obj_rxn == 1.0]\
                .loc[:, ['id', 'flux']].reset_index(drop=True)

            # concatenate DataFrames
            concat_df = pd.concat([met_in_df, met_out_df, obj_df], axis=1)

            del met_in_df, met_out_df, obj_df

            # generate column names
            concat_df.columns = pd.MultiIndex.from_tuples(
                [tuple(c.split('-')) for c in column_names]
            )

        else:
            column_names = ['IN_FLUXES-ID', 'IN_FLUXES-FLUX',
                            'OUT_FLUXES-ID', 'OUT_FLUXES-FLUX',
                            'OBJECTIVES-ID', 'OBJECTIVES-FLUX']

            # generate separate DataFrames
            met_in_df = flux_df[flux_df.is_input == 1.0]\
                .loc[:, ['id', 'flux']].reset_index(drop=True)

            met_out_df = flux_df[flux_df.is_input == 0.0]\
                .loc[:, ['id', 'flux']].reset_index(drop=True)

            obj_df = flux_df[flux_df.is_obj_rxn == 1.0]\
                .loc[:, ['id', 'flux']].reset_index(drop=True)

            # concatenate DataFrames
            concat_df = pd.concat([met_in_df, met_out_df, obj_df], axis=1)

            del met_in_df, met_out_df, obj_df

            # generate column names
            concat_df.columns = pd.MultiIndex.from_tuples(
                [tuple(c.split('-')) for c in column_names]
            )

        return concat_df

    def _to_table(self):
        """
        Returns
        -------
        A string of the summary table.

        """
        return self.to_frame().to_string(header=True, index=False, na_rep='',
                                         float_format=self.float_format,
                                         sparsify=False, justify='center')



from os.path import join

import pytest

import cobra
from cobra.io import read_sbml_model, write_sbml_model


def test_notes(tmp_path):
    """Testing if model notes are written in SBML"""
    path_to_file = join(str(tmp_path), "model_notes.xml")

    # making a minimal cobra model to test notes
    model = cobra.Model("e_coli_core")
    model.notes["Remark"] = "...Model Notes..."
    met = cobra.Metabolite("pyr_c", compartment="c")
    model.add_metabolites([met])
    met.notes["Remark"] = "Note with \n newline"
    rxn = cobra.Reaction("R_ATPM")
    model.add_reactions([rxn])
    rxn.notes["Remark"] = "What about me?"
    model.objective_direction = "max"
    model.objective = rxn
    write_sbml_model(model, path_to_file)

    # reading the model back
    model_after_reading = read_sbml_model(path_to_file)
    met_after_reading = model_after_reading.metabolites.get_by_id("pyr_c")
    reaction_after_reading = model_after_reading.reactions.get_by_id("R_ATPM")

    # checking if notes are written to model
    assert model_after_reading.notes["Remark"] == "...Model Notes..."

    # checking notes for metabolite and reaction
    assert met_after_reading.notes["Remark"] == "Note with \n newline"
    assert reaction_after_reading.notes["Remark"] == "What about me?"



# -*- coding: utf-8 -*-

"""Contains function to identify the type of boundary reactions.

This module uses various heuristics to decide whether a boundary reaction
is an exchange, demand or sink reaction. It mostly orientates on the following
paper:

Thiele, I., & Palsson, B. Ø. (2010, January). A protocol for
generating a high-quality genome-scale metabolic reconstruction.
Nature protocols. Nature Publishing Group.
http://doi.org/10.1038/nprot.2009.203
"""

import logging
from collections import Counter

import pandas as pd

from cobra.medium.annotations import compartment_shortlist, excludes, sbo_terms


LOGGER = logging.getLogger(__name__)


def find_external_compartment(model):
    """Find the external compartment in the model.

    Uses a simple heuristic where the external compartment should be the one
    with the most exchange reactions.

    Arguments
    ---------
    model : cobra.Model
        A cobra model.

    Returns
    -------
    str
        The putative external compartment.
    """
    if model.boundary:
        counts = pd.Series(tuple(r.compartments)[0] for r in model.boundary)
        most = counts.value_counts()
        most = most.index[most == most.max()].to_series()
    else:
        most = None
    like_external = compartment_shortlist["e"] + ["e"]
    matches = pd.Series([co in like_external for co in model.compartments],
                        index=model.compartments)

    if matches.sum() == 1:
        compartment = matches.index[matches][0]
        LOGGER.info("Compartment `%s` sounds like an external compartment. "
                    "Using this one without counting boundary reactions" %
                    compartment)
        return compartment
    elif most is not None and matches.sum() > 1 and matches[most].sum() == 1:
        compartment = most[matches[most]][0]
        LOGGER.warning("There are several compartments that look like an "
                       "external compartment but `%s` has the most boundary "
                       "reactions, so using that as the external "
                       "compartment." % compartment)
        return compartment
    elif matches.sum() > 1:
        raise RuntimeError("There are several compartments (%s) that look "
                           "like external compartments but we can't tell "
                           "which one to use. Consider renaming your "
                           "compartments please.")

    if most is not None:
        return most[0]
        LOGGER.warning("Could not identify an external compartment by name and"
                       " choosing one with the most boundary reactions. That "
                       "might be complete nonsense or change suddenly. "
                       "Consider renaming your compartments using "
                       "`Model.compartments` to fix this.")
    # No info in the model, so give up
    raise RuntimeError("The heuristic for discovering an external compartment "
                       "relies on names and boundary reactions. Yet, there "
                       "are neither compartments with recognized names nor "
                       "boundary reactions in the model.")


def is_boundary_type(reaction, boundary_type, external_compartment):
    """Check whether a reaction is an exchange reaction.

    Arguments
    ---------
    reaction : cobra.Reaction
        The reaction to check.
    boundary_type : str
        What boundary type to check for. Must be one of
        "exchange", "demand", or "sink".
    external_compartment : str
        The id for the external compartment.

    Returns
    -------
    boolean
        Whether the reaction looks like the requested type. Might be based
        on a heuristic.
    """
    # Check if the reaction has an annotation. Annotations dominate everything.
    sbo_term = reaction.annotation.get("sbo", "")
    if isinstance(sbo_term, list):
        sbo_term = sbo_term[0]
    sbo_term = sbo_term.upper()

    if sbo_term == sbo_terms[boundary_type]:
        return True
    if sbo_term in [sbo_terms[k] for k in sbo_terms if k != boundary_type]:
        return False

    # Check if the reaction is in the correct compartment (exterior or inside)
    correct_compartment = external_compartment in reaction.compartments
    if boundary_type != "exchange":
        correct_compartment = not correct_compartment

    # Check if the reaction has the correct reversibility
    rev_type = True
    if boundary_type == "demand":
        rev_type = not reaction.reversibility
    elif boundary_type == "sink":
        rev_type = reaction.reversibility

    return (reaction.boundary and not
            any(ex in reaction.id for ex in excludes[boundary_type]) and
            correct_compartment and rev_type)


def find_boundary_types(model, boundary_type, external_compartment=None):
    """Find specific boundary reactions.

    Arguments
    ---------
    model : cobra.Model
        A cobra model.
    boundary_type : str
        What boundary type to check for. Must be one of
        "exchange", "demand", or "sink".
    external_compartment : str or None
        The id for the external compartment. If None it will be detected
        automatically.

    Returns
    -------
    list of cobra.reaction
        A list of likely boundary reactions of a user defined type.
    """
    if not model.boundary:
        LOGGER.warning("There are no boundary reactions in this model. "
                       "Therefore specific types of boundary reactions such "
                       "as 'exchanges', 'demands' or 'sinks' cannot be "
                       "identified.")
        return []
    if external_compartment is None:
        external_compartment = find_external_compartment(model)
    return model.reactions.query(
        lambda r: is_boundary_type(r, boundary_type, external_compartment))



# -*- coding: utf-8 -*-

"""Test functionalities provided by yaml.py"""

from __future__ import absolute_import

import json
from os.path import join

import pytest
from ruamel.yaml import YAML

import cobra.io as cio
from cobra.test.test_io.conftest import compare_models


def test_load_yaml_model(data_directory, mini_model):
    """Test the reading of YAML model."""
    yaml_model = cio.load_yaml_model(join(data_directory, "mini.yml"))
    assert compare_models(mini_model, yaml_model) is None


@pytest.mark.xfail(reason="schema outdated")
def test_save_yaml_model(tmpdir, mini_model):
    jsonschema = pytest.importorskip("jsonschema")
    """Test the writing of YAML model."""
    output_file = tmpdir.join("mini.yml")
    cio.save_yaml_model(mini_model, output_file.strpath, sort=True)
    # validate against schema
    yaml = YAML(typ="unsafe")
    with open(output_file.strpath, "r") as infile:
        yaml_to_dict = yaml.load(infile)
    dict_to_json = json.dumps(yaml_to_dict)
    loaded = json.loads(dict_to_json)
    assert jsonschema.validate(loaded, cio.json.json_schema)



# -*- coding: utf-8 -*-

"""Helper functions for all flux analysis methods."""

from __future__ import absolute_import

import logging


LOGGER = logging.getLogger(__name__)


def normalize_cutoff(model, zero_cutoff=None):
    """Return a valid zero cutoff value."""
    if zero_cutoff is None:
        return model.tolerance
    else:
        if zero_cutoff < model.tolerance:
            raise ValueError(
                "The chosen zero cutoff cannot be less than the model's "
                "tolerance value."
            )
        else:
            return zero_cutoff



# -*- coding: utf-8 -*-

"""Define the ReactionSummary class."""

from __future__ import absolute_import

from operator import attrgetter

import pandas as pd
from six import iterkeys, itervalues

from cobra.core.summary import Summary


class ReactionSummary(Summary):
    """
    Define the reaction summary.

    Attributes
    ----------
    reaction: cobra.Reaction
        The reaction to summarize.

    See Also
    --------
    Summary : Parent that defines further attributes.
    MetaboliteSummary
    ModelSummary

    """

    def __init__(self, reaction, model, **kwargs):
        """
        Initialize a metabolite summary.

        Parameters
        ----------
        reaction: cobra.Reaction
            The reaction object whose summary we intend to get.
        model : cobra.Model
            The metabolic model in which to generate a reaction summary.

        Other Parameters
        ----------------
        kwargs :
            Further keyword arguments are passed on to the parent class.

        See Also
        --------
        Summary : Parent that has further default parameters.
        MetaboliteSummary
        ModelSummary

        """
        super(ReactionSummary, self).__init__(model=model, **kwargs)
        self.reaction = reaction

    def _generate(self):
        """
        Returns
        -------
        rxn_summary: pandas.DataFrame
            The DataFrame of reaction summary data.

        """
        if self.names:
            emit = attrgetter('name')
        else:
            emit = attrgetter('id')

        data = {
            'GENES_ID': [emit(gene) for gene in self.reaction.genes],
            'METABOLITES_ID':
            [emit(met) for met in iterkeys(self.reaction.metabolites)],
            'METABOLITES_STOICHIOMETRY':
            [met for met in itervalues(self.reaction.metabolites)],
            'METABOLITES_COMPARTMENT':
            [met.compartment for met in iterkeys(self.reaction.metabolites)]
        }

        rxn_summary = pd.DataFrame.from_dict(data, orient='index')\
                                  .T.fillna(value=pd.np.nan)

        rxn_summary.columns = pd.MultiIndex.from_tuples(
            [tuple(c.split('_')) for c in rxn_summary.columns])

        return rxn_summary

    def to_frame(self):
        """
        Returns
        -------
        A pandas.DataFrame of the summary.

        """
        return self._generate()

    def _to_table(self):
        """
        Returns
        -------
        A string of the summary table.

        """
        return self.to_frame().to_string(header=True, index=False, na_rep='',
                                         sparsify=False, justify='center')



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from cobra.sampling.hr_sampler import HRSampler, shared_np_array, step
from cobra.sampling.achr import ACHRSampler
from cobra.sampling.optgp import OptGPSampler
from cobra.sampling.sampling import sample



# -*- coding: utf-8 -*-

"""Define the Summary class."""

from __future__ import absolute_import, division

from warnings import warn

from cobra.flux_analysis.helpers import normalize_cutoff


class Summary(object):
    """
    Define the abstract base summary.

    Attributes
    ----------
    model : cobra.Model
        The metabolic model in which to generate a summary description.
    solution : cobra.Solution
        A solution that matches the given model.
    threshold : float, optional
        Threshold below which fluxes are not reported.
    fva : pandas.DataFrame, optional
        The result of a flux variability analysis (FVA) involving reactions of
        interest if an FVA was requested.
    names : bool
        Whether or not to use object names rather than identifiers.
    float_format : callable
        Format string for displaying floats.

    Methods
    -------
    to_frame
        Return a data frame representation of the summary.

    """

    def __init__(
            self,
            model,
            solution=None,
            threshold=None,
            fva=None,
            names=False,
            float_format="{:.3G}".format,
            **kwargs
    ):
        """
        Initialize a summary.

        Parameters
        ----------
        model : cobra.Model
            The metabolic model in which to generate a summary description.
        solution : cobra.Solution, optional
            A previous model solution to use for generating the summary. If
            None, the summary method will resolve the model.  Note that the
            solution object must match the model, i.e., changes to the model
            such as changed bounds, added or removed reactions are not taken
            into account by this method (default None).
        threshold : float, optional
            Threshold below which fluxes are not reported. May not be smaller
            than the model tolerance (default None).
        fva : pandas.DataFrame or float, optional
            Whether or not to include flux variability analysis in the output.
            If given, fva should either be a previous FVA solution matching the
            model or a float between 0 and 1 representing the fraction of the
            optimum objective to be searched (default None).
        names : bool, optional
            Emit reaction and metabolite names rather than identifiers (default
            False).
        float_format : callable, optional
            Format string for floats (default ``'{:3G}'.format``).

        Other Parameters
        ----------------
        kwargs :
            Further keyword arguments are passed on to the parent class.

        """
        super(Summary, self).__init__(**kwargs)
        self.model = model
        self.solution = solution
        self.threshold = normalize_cutoff(self.model, threshold)
        self.fva = fva
        self.names = names
        self.float_format = float_format

    def _generate(self):
        """Generate the summary for the required cobra object.

        This is an abstract method and thus the subclass needs to
        implement it.

        """
        raise NotImplementedError("This method needs to be implemented by the "
                                  "subclass.")

    def _process_flux_dataframe(self, flux_dataframe):
        """Process a flux DataFrame for convenient downstream analysis.

        This method removes flux entries which are below the threshold and
        also adds information regarding the direction of the fluxes. It is used
        in both ModelSummary and MetaboliteSummary.

        Parameters
        ----------
        flux_dataframe: pandas.DataFrame
            The pandas.DataFrame to process.

        Returns
        -------
        A processed pandas.DataFrame.

        """

        abs_flux = flux_dataframe['flux'].abs()
        flux_threshold = self.threshold * abs_flux.max()

        # Drop unused boundary fluxes i.e., fluxes below threshold
        if self.fva is None:
            flux_dataframe = \
                flux_dataframe.loc[abs_flux >= flux_threshold, :].copy()
        else:
            flux_dataframe = (
                flux_dataframe
                .loc[(abs_flux >= flux_threshold) |
                     (flux_dataframe['fmin'].abs() >= flux_threshold) |
                     (flux_dataframe['fmax'].abs() >= flux_threshold), :]
                .copy()
                )

        # Make all fluxes positive while maintaining proper direction
        if self.fva is None:
            # add information regarding direction
            flux_dataframe['is_input'] = (flux_dataframe['flux'] >= 0)
            # make the fluxes absolute
            flux_dataframe['flux'] = flux_dataframe['flux'].abs()
            # sort the values
            flux_dataframe.sort_values(by=['is_input', 'flux', 'id'],
                                       ascending=[False, False, True],
                                       inplace=True)
        else:

            def get_direction(row):
                """Decide whether or not to reverse a flux to make it
                positive."""

                if row.flux < 0:
                    sign = -1
                elif row.flux > 0:
                    sign = 1
                elif (row.fmax > 0) & (row.fmin <= 0):
                    sign = 1
                elif (row.fmax < 0) & (row.fmin >= 0):
                    sign = -1
                elif ((row.fmax + row.fmin) / 2) < 0:
                    sign = -1
                else:
                    sign = 1

                return sign

            # get a sign DataFrame to use as a pseudo-mask
            sign = flux_dataframe.apply(get_direction, axis=1)

            flux_dataframe['is_input'] = (sign == 1)

            flux_dataframe.loc[:, ['flux', 'fmin', 'fmax']] = (
                flux_dataframe.loc[:, ['flux', 'fmin', 'fmax']]
                .multiply(sign, axis=0)
                .astype('float')
            )

            flux_dataframe.sort_values(by=['is_input', 'flux', 'fmax', 'fmin',
                                           'id'],
                                       ascending=[False, False, False, False,
                                                  True],
                                       inplace=True)

        return flux_dataframe

    def to_frame(self):
        """Generate a pandas DataFrame.

        This is an abstract method and thus the subclass needs to
        implement it.

        """
        raise NotImplementedError("This method needs to be implemented by the "
                                  "subclass.")

    def _to_table(self):
        """Generate a pretty-print table.

        This is an abstract method and thus the subclass needs to
        implement it.

        """
        raise NotImplementedError("This method needs to be implemented by the "
                                  "subclass.")

    def __str__(self):
        if self.float_format is not None:
            warn("Setting float_format to anything other than None "
                 "will cause nan to be present in the output.")
        return self._to_table()

    def _repr_html_(self):
        return self.to_frame()._repr_html_()
